#-*- coding:utf-8 -*-
"""
Module containing Exceptions classes.
"""

class RPCError(Exception):
    """
    Generic RPC Exception.
    """

class TimeoutError(Exception):
    """
    Dummy TimeoutError.
    """



#-*- coding:utf-8 -*-
"""
Module which setup the logging using a yaml file.
"""
import logging
import logging.config
import os
import yaml

def setup_logging(app, disable_existing_loggers=True):
    """
    Setup the logging using logging.yaml.
    :param app: The app which setups the logging. Used for the log's filename and for the log's name.
    :type app: str
    :param disable_existing_loggers: If False, loggers which exist when this call is made are left enabled.
    :type disable_existing_loggers: bool
    :returns: None
    """
    conf = yaml.load(open(os.path.join(os.path.dirname(os.path.abspath(__file__)), 'logging.yaml'), 'r'))
    conf['disable_existing_loggers'] = disable_existing_loggers
    conf['loggers'][app] = conf['loggers'].pop('__name__')
    logging.config.dictConfig(conf)



#-*- coding:utf-8 -*-
"""
Module containing the class decorators:
 * publish()
 * consume()
"""
import logging
import farine.discovery
from farine.amqp.publisher import Publisher
from farine.amqp.consumer import Consumer

LOGGER = logging.getLogger(__name__)

def publish(exchange=None, routing_key=None):
    def wrapper(method):
        def subwrapper(self, *args, **kwargs):
            if not hasattr(self, 'publish'):
                name = getattr(self, 'service', self.__class__.__name__.lower())
                _exchange = exchange or kwargs.get('exchange') or name
                _routing_key = routing_key or kwargs.get('routing_key') or name
                for kw in ['exchange', 'routing_key']:
                    if kwargs.get(kw):
                        kwargs.pop(kw)
                publish = Publisher(_exchange, _routing_key, service=name)
            return method(self, publish, *args, **kwargs)
        return subwrapper
    return wrapper

def consume(*args, **kwargs):
    def wrapper(method):
        farine.discovery.ENTRYPOINTS.append((Consumer, method, args, kwargs))
        def subwrapper(self, *args, **kwargs):
            LOGGER.info('Received message : %s %s , call %s.', args, kwargs, method.__name__)
            return method(self, *args, **kwargs)
        return subwrapper
    return wrapper



#-*- coding:utf-8 -*-
"""
| Module containing everything related to service management:
| Register services to launch, start them.


|TODO: J'en ai vu de la merde, mais de la merde comme ca. CLEANUP.
"""
import importlib
import inspect
import logging
import sys
import gevent.threadpool
LOGGER = logging.getLogger(__name__)


ENTRYPOINTS = [] #List of a tuple: (Entrypoint, callback, args, kwargs)

def import_module(module):
    """
    | Given a module `service`, try to import it.
    | It will autodiscovers all the entrypoints
    | and add them in `ENTRYPOINTS`.

    :param module: The module's name to import.
    :type module: str
    :rtype: None
    :raises ImportError: When the service/module to start is not found.
    """
    try:
        __import__('{0}.service'.format(module))
    except ImportError:
        LOGGER.error('No module/service found. Quit.')
        sys.exit(0)

def import_models(module):
    """
    | Given a module `service`, try to import its models module.

    :param module: The module's name to import the models.
    :type module: str
    :rtype: list
    :returns: all the models defined.
    """
    try:
        module = importlib.import_module('{0}.models'.format(module))
        md = module.__dict__
    except ImportError:
        return []
    else:
        return [md[c] for c in md if (isinstance(md[c], type) and md[c].__module__ == module.__name__)]

def start():
    """
    | Start all the registered entrypoints
    | that have been added to `ENTRYPOINTS`.

    :rtype: None
    """
    pool = gevent.threadpool.ThreadPool(len(ENTRYPOINTS))
    for entrypoint, callback, args, kwargs in ENTRYPOINTS:
        cname = callback.__name__
        #1. Retrieve the class which owns the callback
        for name, klass in inspect.getmembers(sys.modules[callback.__module__], inspect.isclass):
            if hasattr(klass, cname):
                service_name = name.lower()
                break
        #2.Start the entrypoint
        callback = getattr(klass(), cname)
        kwargs.update({'service':service_name, 'callback':callback, 'callback_name': cname})
        LOGGER.info('Start service %s[%s].', service_name.capitalize(), cname)
        obj = entrypoint(*args, **kwargs)
        pool.spawn(obj.start, *args, **kwargs)
    pool.join()
    return True



#-*- coding:utf-8 -*-
"""
Utils containing useful tools.
"""
import contextlib
import signal
import logging
import farine.exceptions as exceptions

LOGGER = logging.getLogger(__name__)

@contextlib.contextmanager
def suppress(*exceptions):
    """
    | Log and ignore `exceptions`.
    | If another exception occurs, it will be raised.

    :param exceptions: Exceptions to handle.
    :type exceptions: tuple
    :rtype: None
    """
    try:
        yield
    except exceptions as exc:
        LOGGER.error('Error occured, ignore it :%s.', exc)



class Timeout(object):
    """
    Dummy timeout context manager.
    """

    def __init__(self, seconds):
        """
        :param seconds: the timeout must be specified in seconds.
        :type seconds: int
        """
        self.seconds = int(seconds) if seconds else None

    def handler(self, signum, frame):
        """
        Handler of the timeout.
        """
        raise exceptions.TimeoutError('{0} seconds reached.'.format(self.seconds))

    def __enter__(self):
        """
        Create a timeout if seconds is specified.
        """
        if self.seconds:
            signal.signal(signal.SIGALRM, self.handler)
            signal.alarm(self.seconds)

    def __exit__(self, type, value, traceback):
        """
        Deactivate the timeout.
        """
        if self.seconds:
            signal.alarm(0)



#-*- coding:utf-8 -*-
from fixtures import *
import farine.settings


def test_defaults():
    assert hasattr(farine.settings, 'publish')
    assert farine.settings.publish.has_key('amqp_uri')
    ##
    assert farine.settings.publish['type'] == 'direct'
    assert farine.settings.publish['durable'] == True
    assert farine.settings.publish['delivery_mode'] == 2
    assert farine.settings.publish['retry'] == True
    assert farine.settings.publish['retry_policy'] == {'max_retries': 5}
    assert farine.settings.publish['serializer'] == 'json'



#-*- coding:utf-8 -*-
import json
import farine.discovery
import sel.serializers
from peewee import Model as PModel
from peewee import *
from playhouse.shortcuts import model_to_dict
try:
    from playhouse.postgres_ext import *
except ImportError:
    pass


class Model(PModel):
    """
    Base model that extends the peewee one.
    """

    def to_json(self):
        """
        Convert a model into a json using the playhouse shortcut.
        """
        return json.dumps(model_to_dict(self), cls=sel.serializers.JsonEncoder)

class Meta:
    database = None

def setup(settings):
    """
    Setup the database connection.
    """
    connector = settings.get('db_connector')
    if connector == 'postgres':
        from playhouse.pool import PooledPostgresqlExtDatabase
        return PooledPostgresqlExtDatabase(settings['db_name'],
                           user=settings['db_user'],
                           password=settings['db_password'],
                           host=settings['db_host'],
                           port=settings.get('db_port'),
                           max_connections=settings.get('db_max_conn'),
                           stale_timeout=settings.get('db_stale_timeout'),
                           timeout=settings.get('db_timeout'),
                           register_hstore=False)

def init(module, db):
    """
    Initialize the models.
    """
    for model in farine.discovery.import_models(module):
        model._meta.database = db



#-*- coding:utf-8 -*-
import json
from fixtures import *
import farine.rpc as rpc

def test_rpc_decorator_call_ok(rpc_server_factory, rpc_client_factory, rabbitmq_proc, rabbitmq):
    """
    Test that when a client call the RPC server,
    its message is processed.
    """
    clear_rabbitmq(rabbitmq_proc, rabbitmq)
    server = rpc_server_factory('something')
    client = rpc_client_factory()
    assert client.call() == True

def test_rpc_decorator_call_timeout(rpc_client_factory, rabbitmq_proc, rabbitmq):
    """
    Test that when a client call a non existing RPC server,
    it timeouts.
    """
    clear_rabbitmq(rabbitmq_proc, rabbitmq)
    client = rpc_client_factory()
    with pytest.raises(exceptions.RPCError):
        client.call()

def test_rpc_decorator_call_exception(rpc_server_factory, rpc_client_factory, rabbitmq_proc, rabbitmq):
    """
    Test that when the server raises an error, it's propagated
    to the client.
    """
    clear_rabbitmq(rabbitmq_proc, rabbitmq)
    server = rpc_server_factory('exception')
    client = rpc_client_factory()
    assert client.call_exception() == False

def test_rpc_class_call_ok(rpc_server_factory, rabbitmq_proc, rabbitmq):
    """
    Test that rpc.client.Client() works like rpc.client() decorator:
    its message is processed.
    """
    clear_rabbitmq(rabbitmq_proc, rabbitmq)
    server = rpc_server_factory('something')
    client = rpc.Client('server')
    assert client.something() == True

def test_rpc_class_call_timeout(rabbitmq_proc, rabbitmq):
    """
    Test that rpc.client.Client() works like rpc.client() decorator:
    its timeouts.
    """
    clear_rabbitmq(rabbitmq_proc, rabbitmq)
    client = rpc.Client('server', 2)
    with pytest.raises(exceptions.RPCError) as err:
        client.call()
    assert 'timeout' in err.value.message

def test_rpc_long_call(rpc_server_factory, rabbitmq_proc, rabbitmq):
    """
    Test that rpc.client.Client() timeouts
    then the server received the message, but hasn't reply yet.
    """
    clear_rabbitmq(rabbitmq_proc, rabbitmq)
    client = rpc.Client('server', 10)
    with pytest.raises(exceptions.RPCError) as err:
        client.slow()
    assert 'timeout' in err.value.message

def test_rpc_class_call_exception(rpc_server_factory, rabbitmq_proc, rabbitmq):
    """
    Test that rpc.client.Client() works like rpc.client() decorator:
    the server raises an error.
    """
    clear_rabbitmq(rabbitmq_proc, rabbitmq)
    server = rpc_server_factory('exception')
    client = rpc.Client('server')
    with pytest.raises(exceptions.RPCError):
        client.exception()

def test_rpc_streaming_class_call_ok(rpc_server_factory, rabbitmq_proc, rabbitmq):
    """
    Test that rpc.client.Client() works like rpc.client() decorator with streaming:
    its message is processed.
    """
    clear_rabbitmq(rabbitmq_proc, rabbitmq)
    server = rpc_server_factory('stream')
    client = rpc.Client('server', 60)
    response = [i for i in client.stream(__stream__=True)]
    assert len(response) == 3



#-*- coding:utf-8 -*-
"""
The consumer module manage all the consumer's messages workflow:
    - Connection to the server
    - Consume the message from an exchange
    - Error handler (reconnection, retry, etc.)
"""
import contextlib
import cProfile
import pstats
import StringIO
import uuid
from kombu import Connection, Exchange, Queue
from kombu.pools import producers
from kombu.mixins import ConsumerMixin
from farine.mixins import EntryPointMixin
import farine.settings

class Consumer(ConsumerMixin, EntryPointMixin):
    """
    Consumer generic class.
    """
    prefetch_count = None
    exclusive = False
    auto_delete = False
    auto_generated = False
    routing_key_format = None

    def __init__(self, *args, **kwargs):#pylint:disable=unused-argument
        """
        :param service: The service's name which consume.
        :type service: str
        :param queue_name: The name of queue. Optional, default to `service` value.
        :type queue_name: None, str
        :param exhange: The exchange's name to consume from.
        :type exchange: str
        :param exchange_type: The exchange's type. Default to 'direct'.
        :type exchange_type: str
        :param routing_key: The routing key to read the message.
        :type routing_key: str
        :param callback: The callback to call when receiving a message.
        :type callback: object
        :rtype: None
        """
        self.service = kwargs.get('service')
        kwargs.setdefault('exchange', self.service)
        if self.auto_generated:
            self.queue_name = self.routing_key = uuid.uuid4().hex
        elif self.routing_key_format:
            self.queue_name = self.routing_key = self.routing_key_format.format(**kwargs)
        else:
            self.queue_name = self.routing_key = kwargs.get('routing_key') or self.service
        self.settings = getattr(farine.settings, self.service)
        if not self.callback:
            self.callback = kwargs.get('callback')
        exchange_type = kwargs.pop('exchange_type', 'direct')
        self.exchange = Exchange(kwargs.pop('exchange'),
                                 type=exchange_type,
                                 durable=self.settings['durable'],
                                 auto_declare=self.settings['auto_declare'],
                                 delivery_mode=self.settings['delivery_mode'])
        self.queue = Queue(self.queue_name,
                           exchange=self.exchange,
                           routing_key=self.routing_key,
                           exclusive=self.settings.get('exclusive', self.exclusive),
                           auto_delete=self.settings.get('auto_delete', self.auto_delete),
                           durable=self.settings['durable'],
                           auto_declare=self.settings['auto_declare'])
        self.connection = Connection(self.settings['amqp_uri'])

    def get_consumers(self, _Consumer, channel):
        """
        | ConsumerMixin requirement.
        | Get the consumers list.

        :returns: All the consumers.
        :rtype: list.
        """
        return [_Consumer(queues=[self.queue(channel)], callbacks=[self.main_callback], prefetch_count=self.prefetch_count)]


    @contextlib.contextmanager
    def debug(self, body, message):#pylint:disable=arguments-differ,unused-argument
        """
        | Rewrite/move some code?
        | EntryPointMixin requirement. Context Manager.
        | Check if debug is enabled for the message consumed.
        | If so will run cProfile for this message, and send the result
        | to the `__debug__` queue of the exchange.

        :param callback: the callback of the message consumed.
        :type callback: function
        :rtype: None
        """
        is_debug = body.get('__debug__', False) if isinstance(body, dict) else False
        if not is_debug:
            yield
        else:
            #Start the profiler
            result = StringIO.StringIO()
            profiler = cProfile.Profile()
            profiler.enable()
            yield
            #Close the profiler
            profiler.disable()
            profiler.create_stats()
            stats = pstats.Stats(profiler, stream=result)
            stats.strip_dirs().print_stats()
            debug_message = body.copy()
            debug_message['__debug__'] = result.getvalue()
            #Send the result to the exchange
            with producers[self.connection].acquire(block=True) as producer:
                debug_queue = Queue('debug',
                                    exchange=self.exchange,
                                    routing_key='debug',
                                    durable=True,
                                    auto_declare=True)
                producer.maybe_declare(debug_queue)
                producer.publish(debug_message,
                                 exchange=self.exchange,
                                 declare=[self.exchange],
                                 routing_key='debug')


    def start(self, *args, **kwargs):#pylint:disable=unused-argument
        """
        | Launch the consumer.
        | It can listen forever for messages or just wait for one.

        :param forever: If set, the consumer listens forever. Default to `True`.
        :type forever: bool
        :param timeout: If set, the consumer waits the specified seconds before quitting.
        :type timeout: None, int
        :rtype: None
        :raises socket.timeout: when no message has been received since `timeout`.
        """
        forever = kwargs.get('forever', True)
        timeout = kwargs.get('timeout', None)
        if forever:
            return self.run(timeout=timeout)
        elif timeout:
            next((self.consume(timeout=timeout)), None)
        else:
            next((self.consume(limit=1, timeout=timeout)), None)

    def stop(self):
        """
        Stop to consume.

        :rtype: None
        """
        self.should_stop = True
        self.connection.release()



#-*-coding:utf-8 -*-
"""
Module containing the class decorators:
 * client()
 * method()
"""
import logging
import farine.discovery
from farine.rpc.client import Client
from farine.rpc.server import Server

LOGGER = logging.getLogger(__name__)

def client(service, timeout=None):
    def wrapper(_method):
        def subwrapper(self, *args, **kwargs):
            if not hasattr(self, 'rpc'):
                rpc = Client(service, timeout)
            return _method(self, rpc, *args, **kwargs)
        return subwrapper
    return wrapper

def method(*args, **kwargs):
    def wrapper(_method):
        farine.discovery.ENTRYPOINTS.append((Server, _method, args, kwargs))
        def subwrapper(self, *args, **kwargs):
            LOGGER.info('Received message : %s %s , call %s.', args, kwargs, _method.__name__)
            return _method(self, *args, **kwargs)
        return subwrapper
    return wrapper



#-*- coding:utf-8 -*-
import json
from fixtures import *

def test_sse_call_ok(sse_server_ok, sse_client_factory):
    """
    Test that when a client call the stream
    an event is received.
    """
    client = sse_client_factory()
    worker_client = stream.SSEConsumer(service='streamsseclient', callback=client.event)
    assert client.messages_consumed == 0
    worker_client.start(limit=1)
    assert client.messages_consumed == 1



#-*- coding:utf-8 -*-
import multiprocessing
import time
import pytest
import farine.amqp
import farine.log
import farine.exceptions as exceptions
import farine.rpc as rpc
import farine.stream as stream
import farine.settings
import kombu.log
import requests
import requests_mock
from rabbitpy import Exchange, Queue, Message
from pytest_rabbitmq.factories.client import clear_rabbitmq
from .models import User

@pytest.fixture(autouse=True)
def settings():
    farine.settings.load()

@pytest.fixture(autouse=True)
def logging():
    farine.log.setup_logging(__name__, False)
    kombu.log.setup_logging(loglevel='INFO')

class Publish(object):

    @farine.amqp.publish()
    def publish_dummy(self, publish, message):
        publish(message)
        return True

class Consume(object):
    messages_consumed = 0

    @farine.amqp.consume(exchange='publish', routing_key='publish', forever=False)
    def consume_dummy(self, body, message):
        self.messages_consumed += 1
        return True

@pytest.fixture()
def publisher_factory():
    return Publish

@pytest.fixture()
def consumer_factory():
    return Consume

class Server(object):
    @rpc.method()
    def something(self, *args, **kwargs):
        return True

    @rpc.method()
    def exception(self, *args, **kwargs):
        return i_am_broken

    @rpc.method()
    def slow(self, *args, **kwargs):
        time.sleep(60)
        return True

    @rpc.method()
    def stream(self, *args, **kwargs):
        yield 'a'
        yield 'b'
        yield 'c'

class Model(object):
    @rpc.method()
    def save(self, name):
        with self.db.transaction() as txn:
            User.create(name=name)
            txn.commit()
        return True

    @rpc.method()
    def get(self):
        return User.select().count()

class Client(object):
    @rpc.client('server', 40)
    def call(self, rpc):
        result = rpc.something('un', deux='deux')
        return result

    @rpc.client('server', 40)
    def call_exception(self, rpc):
        try:
            rpc.exception()
            return True
        except exceptions.RPCError:
            return False

    @rpc.client('model', 40)
    def save(self, rpc, name):
        return rpc.save(name)

    @rpc.client('model', 40)
    def get(self, rpc):
        return rpc.get()


def _rpc_server(callback_name, service='server'):
    callbacks = {'server': Server, 'model': Model}
    farine.settings.load()
    rpc.Server(service=service, callback_name=callback_name, callback=getattr(callbacks[service](), callback_name)).start()

@pytest.fixture()
def rpc_server_factory(request, rabbitmq_proc, rabbitmq):
    def factory(callback_name, service='server'):
        process = multiprocessing.Process(
            target=lambda:_rpc_server(callback_name, service),
        )
        def cleanup():
            process.terminate()
            clear_rabbitmq(rabbitmq_proc, rabbitmq)
        request.addfinalizer(cleanup)
        process.start()
        time.sleep(5)
        return process
    return factory

class StreamSSEClient(object):
    messages_consumed = 0

    @stream.http()
    def event(self, data):
        self.messages_consumed += 1
        return data

@pytest.fixture()
def sse_client_factory():
    return StreamSSEClient

@pytest.fixture()
def sse_server_ok():
    with requests_mock.mock() as m:
        text = """event: event_stream_detached
data: {"remoteAddress":"10.1.0.234","eventType":"event_stream_detached","timestamp":"2017-05-25T17:27:08.373Z"}\n\n"""
        m.get('http://unittest/v2/events', text=text)
        yield

@pytest.fixture()
def rpc_client_factory():
    return Client

@pytest.fixture()
def channel(rabbitmq):
    return rabbitmq.channel()

@pytest.fixture
def queue_factory(rabbitmq, rabbitmq_proc):
    def factory(name, exchange, routing_key):
        channel = rabbitmq.channel()
        exchange = Exchange(channel, exchange, auto_delete=False, durable=True)
        exchange.declare()
        assert exchange.name in rabbitmq_proc.list_exchanges()
        queue = Queue(channel, name, auto_delete=False, durable=True)
        queue.declare()
        queue.bind(exchange, routing_key=routing_key)
        assert name in rabbitmq_proc.list_queues()
        return exchange, queue
    return factory

@pytest.fixture
def message_factory():
    """
    Message factory.
    """
    def factory(service, routing_key, message):
        pub = farine.amqp.Publisher(service, routing_key)
        return pub.send(message)
    return factory

@pytest.fixture()
def amqp_factory(request, rabbitmq_proc, rabbitmq):
    def factory(exchange, routing_key, name, callback, queue_name=None):
        return farine.amqp.Consumer(exchange=exchange, routing_key=routing_key,
            service=name, callback=callback, queue_name=queue_name)
    def cleanup():
        clear_rabbitmq(rabbitmq_proc, rabbitmq)
    request.addfinalizer(cleanup)
    return factory



#-*- coding:utf-8 -*-
"""
Module containing the configuration.

To load the configuration, call the `load()` function.

By default the configuration's filepath */etc/farine.ini* will be use.
To override it, set the environment variable **FARINE_INI**.

The format of the config must be compliant with configparser,
have a [DEFAULT] section and one by service.


Example:

| [DEFAULT]
| amqp_uri=amqp://user:password@localhost:5672/vhost

| [taskstatus]
| enabled = true
"""
import os
import ConfigParser

CONF_PATH = os.environ.get('FARINE_INI', '/etc/farine.ini')
DEFAULTS = {
    'type': 'direct', #Exchange's type : `direct`, `topic`, `broadcast`
    'durable': True, #Does the exchange still exist after the AMQP server restart.
    'auto_declare': True, #Does the queue auto declare itself.
    'delivery_mode': 2, #How the messages are stored in the server. Transient=>1 or Persistent=>2
    'retry': True, #Retry sending message or declaring the exchange if the connection is lost.
    'retry_policy' : {'max_retries': 5}, #Retry's policy.
    'serializer': 'json', #The serializer used to encode the message.
}

def load():
    """
    | Load the configuration file.
    | Add dynamically configuration to the module.

    :rtype: None
    """
    config = ConfigParser.RawConfigParser(DEFAULTS)
    config.readfp(open(CONF_PATH))
    for section in config.sections():
        globals()[section] = {}
        for key, val in config.items(section):
            globals()[section][key] = val



#-*- coding:utf-8 -*-
"""
Module containing a consumer
to a stream HTTP Connection.
"""
import contextlib
import farine.settings
import farine.utils as utils
from farine.mixins import EntryPointMixin
import sseclient

class SSEConsumer(EntryPointMixin):
    """
    HTTP SSE consumer.
    """

    def __init__(self, *args, **kwargs):#pylint:disable=unused-argument
        """
        :param service: The service's name which consume the stream.
        :type service: str
        :param callback: The callback to call when receiving a message.
        :type callback: object
        :rtype: None
        """
        self.callback = kwargs.get('callback')
        self.service = kwargs.get('service')
        self.settings = getattr(farine.settings, self.service)
        self.endpoint = kwargs.get('endpoint', self.settings['endpoint'])
        self.headers = kwargs.get('headers', {'Accept':'text/event-stream'})
        self.stream = None


    @contextlib.contextmanager
    def debug(self, data):#pylint:disable=arguments-differ,unused-argument
        """
        Add a debug method.
        """
        yield

    def run(self, limit=None, timeout=None):
        """
        Consume the event stream.
        :param timeout: Duration of the connection timeout.
        :type timeout: int
        :param limit: Number of events to consume.
        :type limit: int
        :rtype: None
        """
        counter = 0
        self.stream = sseclient.SSEClient(self.endpoint)
        while True:
            with utils.Timeout(timeout):
                try:
                    event = next(self.stream)
                except StopIteration:
                    continue
                else:
                    if not event.data:
                        continue
                    self.main_callback(event.data)
                    counter += 1
                    if limit and counter >= limit:
                        return

    def start(self, *args, **kwargs):#pylint:disable=unused-argument
        """
        | Launch the SSE consumer.
        | It can listen forever for messages or just wait for one.

        :param limit: If set, the consumer listens for a limited number of events.
        :type limit: int
        :param timeout: If set, the consumer listens for an event for a limited time.
        :type timeout: int
        :rtype: None
        """
        limit = kwargs.get('limit', None)
        timeout = kwargs.get('timeout', None)
        self.run(limit=limit, timeout=timeout)

    def stop(self):
        """
        Stop the consumer.
        :rtype: None
        """



#-*- coding:utf-8 -*-
import time
import pytest
import farine.utils
import farine.exceptions
from fixtures import *


def test_suppress_silent():
    with farine.utils.suppress(ValueError):
        raise ValueError

def test_suppress_raises():
    with pytest.raises(ValueError) as e:
        with farine.utils.suppress((RuntimeError, ZeroDivisionError)):
            raise ValueError
    assert e.type == ValueError

def test_timeout_ok():
    with farine.utils.Timeout(5):
        time.sleep(0.1)

def test_timeout_none_ok():
    with farine.utils.Timeout(None):
        time.sleep(0.1)

def test_timeout_raise():
    with pytest.raises(farine.exceptions.TimeoutError) as e:
        with farine.utils.Timeout(1):
            time.sleep(2)
    assert e.type == farine.exceptions.TimeoutError



#-*- coding:utf-8 -*-
"""
Module containing all the farine mixins.
"""
import abc
import contextlib
import farine.connectors.sql as sql

class EntryPointMixin(object):
    """
    | Entry point Mixin.
    | All services that are listening for an event(like a consumer, an http server, ...)
    | are called `Entry Point` and then must inherit from it.
    | They will implement debug mode, monitoring, etc.
    """
    __metaclass__ = abc.ABCMeta

    callback = None

    def main_callback(self, *args, **kwargs):
        """
        Main callback called when an event is received from an entry point.

        :returns: The entry point's callback.
        :rtype: function
        :raises NotImplementedError: When the entrypoint doesn't have the required attributes.
        """
        if not self.callback:
            raise NotImplementedError('Entrypoints must declare `callback`')
        if not self.settings:
            raise NotImplementedError('Entrypoints must declare `settings`')

        self.callback.im_self.db = None

        #1. Start all the middlewares
        with self.debug(*args, **kwargs):
            with self.database():
                #2. `Real` callback
                result = self.callback(*args, **kwargs)#pylint: disable=not-callable
        return result

    @abc.abstractmethod
    def debug(self):
        """
        | The debug implementation for the entry point,
        | Each entry point must have it's own debug logic.
        """

    @contextlib.contextmanager
    def database(self):
        """
        Before the callback is called, initialize the database if needed.
        :rtype: None
        """
        #1. Initialize
        self.callback.im_self.db = sql.setup(self.settings)
        if self.callback.im_self.db:
            module = '.'.join(self.callback.im_self.__module__.split('.')[:-1])
            sql.init(module, self.callback.im_self.db)
            self.callback.im_self.db.connect()
        yield
        #2. Cleanup
        if self.callback.im_self.db:
            self.callback.im_self.db.close()



#-*- coding:utf-8 -*-
import json
from fixtures import *
from farine.amqp import Consumer

def test_no_has_publish(publisher_factory):
    publisher = publisher_factory()
    assert hasattr(publisher, 'publish') ==  False

def test_exchange_created(rabbitmq, publisher_factory, rabbitmq_proc):
    publisher = publisher_factory()
    assert 'publish' not in rabbitmq_proc.list_exchanges()
    assert publisher.publish_dummy({'test':'ok'})
    assert 'publish' in rabbitmq_proc.list_exchanges()
    clear_rabbitmq(rabbitmq_proc, rabbitmq)

def test_message_sent(publisher_factory, queue_factory, rabbitmq_proc, rabbitmq):
    exc, queue = queue_factory('consume', 'publish', 'publish')
    publisher = publisher_factory()
    publisher.publish_dummy({'test':'ok'})
    message = next(queue.consume())
    assert json.loads(message.body) == {"test": "ok"}
    clear_rabbitmq(rabbitmq_proc, rabbitmq)

def test_consumer_read(queue_factory, rabbitmq_proc, rabbitmq, consumer_factory, publisher_factory):
    exc, queue = queue_factory('publish', 'publish', 'publish')
    consumer = consumer_factory()
    worker_consumer = Consumer(exchange='publish', routing_key='publish', service='consume', callback=consumer.consume_dummy)
    publisher_factory().publish_dummy({'test':'ok'})
    #
    worker_consumer.start(forever=False)
    assert 'publish' in rabbitmq_proc.list_queues()
    assert consumer.messages_consumed == 1
    clear_rabbitmq(rabbitmq_proc, rabbitmq)

def test_debug(queue_factory, rabbitmq_proc, rabbitmq, consumer_factory, publisher_factory):
    exc, queue = queue_factory('publish', 'publish', 'publish')
    consumer = consumer_factory()
    worker_consumer = Consumer(exchange='publish', routing_key='publish', service='consume', callback=consumer.consume_dummy)
    publisher_factory().publish_dummy({'test':'ok', '__debug__':True})
    worker_consumer.start(forever=False)
    #Check that the queue debug auto created has one message
    exc, debug = queue_factory('debug', 'publish', 'debug')
    debug_msg = json.loads(next(debug.consume()).body)
    assert '__debug__' in debug_msg.keys()
    clear_rabbitmq(rabbitmq_proc, rabbitmq)



#-*- coding:utf-8 -*-
"""
Entrypoint of the sql connectors.
"""
import argparse
import farine.connectors.sql
import farine.discovery
import farine.settings
import farine.log

farine.log.setup_logging(__name__)

def migrate(module=None):
    """
    Entrypoint to migrate the schema.
    For the moment only create tables.
    """
    if not module:
        #Parsing
        parser = argparse.ArgumentParser()
        parser.add_argument('-s', '--service', type=str, help='Migrate the module', required=True,
                            dest='module')
        args = parser.parse_args()
        module = args.module
    #1. Load settings
    farine.settings.load()
    #2. Load the module
    models = farine.discovery.import_models(module)
    #3. Get the connection
    db = farine.connectors.sql.setup(getattr(farine.settings, module))
    #4. Create tables
    for model in models:
        model._meta.database = db
        model.create_table(fail_silently=True)



#-*- coding:utf-8 -*-
'''
Setup for the baguette messaging.
'''
from setuptools import find_packages, setup


setup(name='baguette-messaging',
      version='0.12',
      description='Baguette messaging framework',
      long_description=open('README.rst').read(),
      url='https://github.com/baguette-io/baguette-messaging/',
      author_email='pydavid@baguette.io',
      keywords=['micro', 'services', 'amqp', 'rpc', 'messaging'],
      classifiers=[
          'Programming Language :: Python',
          'Programming Language :: Python :: 2.7',
          'License :: OSI Approved',
          'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
          'Operating System :: OS Independent',
          'Topic :: Software Development :: Libraries :: Application Frameworks',
      ],
      packages=find_packages(),
      install_requires=[
          'baguette-utils',
          'gevent==1.1.1',
          'kombu==4.0.2',
          'peewee==2.10.1',
          'PyYAML==3.11',
          'sseclient==0.0.18',
      ],
      extras_require={
          'postgres': [
              'psycopg2==2.7.3.1'
          ],
          'testing': [
              'mock',
              'pytest',
              'pytest-postgresql',
              'pytest-rabbitmq',
              'pytest-cov',
              'pylint',
              'rabbitpy',
              'requests-mock',
          ],
          'doc': [
              'Sphinx==1.4.4',
          ],
      },
      entry_points={
          'console_scripts':[
              'farine=farine.main:main',
              'farine-migrate=farine.connectors.sql.entrypoints:migrate',
          ],
      },
      package_data={
          'farine.tests':['farine.ini', 'pytest.ini'],
          'farine.log':['logging.yaml'],
      },
     )



#-*- coding:utf-8 -*-
"""
Entrypoint of the package.
Can start a module here.
"""
import argparse
import kombu.log
import farine.discovery
import farine.settings
import farine.log

farine.log.setup_logging(__name__)
kombu.log.setup_logging()

def main(module=None):
    """
    Entry point.
    """
    if not module:
        #Parsing
        parser = argparse.ArgumentParser()
        parser.add_argument('-s', '--start', type=str, help='Start the module', required=True,
                            dest='module')
        args = parser.parse_args()
        module = args.module
    #1. Load settings
    farine.settings.load()
    #2. Load the module
    farine.discovery.import_module(module)
    #3. Start the module
    farine.discovery.start()



#-*-coding:utf-8 -*-
import farine.connectors.sql.entrypoints
from pytest_postgresql import factories
from fixtures import *

postgresql_proc = factories.postgresql_proc()
postgres = factories.postgresql('postgresql_proc')

def test_postgres_simple(postgres, rpc_server_factory, rpc_client_factory, rabbitmq_proc, rabbitmq):
    """
    Try to create an object and then count them : must succeed.
    """
    clear_rabbitmq(rabbitmq_proc, rabbitmq)
    db = farine.connectors.sql.setup(farine.settings.model)
    db.execute_sql('CREATE TABLE "user" ("id" SERIAL NOT NULL PRIMARY KEY, "name" VARCHAR(255) NOT NULL)')
    db.close()
    server1 = rpc_server_factory('save', 'model')
    server2 = rpc_server_factory('get', 'model')
    client = rpc_client_factory()
    #
    assert client.save('david') == True
    assert client.get() == 1
    #
    db.connect()
    assert db.execute_sql('SELECT name FROM "user"').fetchall() == [('david',)]
    db.close()



#-*- coding:utf-8 -*-
#pylint:disable=attribute-defined-outside-init,method-hidden
"""
RPC over AMQP implementation: client side.
"""
import Queue
import uuid
import traceback

import farine.amqp
import farine.exceptions as exceptions

class Client(farine.amqp.Consumer):
    """
    RPC Client which is
    an AMQP publisher then consumer.
    """
    prefetch_count = 1
    auto_generated = True
    #TODO: delete the queue when the rpc call is done.
    #auto_delete = True
    #exclusive = True

    def __init__(self, service, timeout=None):
        super(Client, self).__init__(service, service, service=service)
        self.timeout = timeout
        self.remote = None
        self.running = True
        self.results = Queue.Queue()

    def callback(self, result, message):
        """
        Method called automatically when a message is received.
        This will be executed before exiting self.start().
        """
        self.results.put(result)
        message.ack()

    def __wrap_rpc__(self, *args, **kwargs):
        """
        | Wrapper for our RPC method:
        | if it's a non streaming call, then we return the next and only one element of the generator.
        | Otherwise we returns the generator.
        """
        stream = kwargs.get('__stream__', False)
        result = self.__rpc__(*args, **kwargs)
        if stream:
            return result
        return next(result)


    def __rpc__(self, *args, **kwargs):
        """
        RPC call logic.
        There are two types of rpc calls:
        Streaming and basic.
        We don't set the auto_delete flag to the queue because of the streaming call.
        """
        self.correlation_id = uuid.uuid4().hex
        message = {'args': args,
                   'kwargs': kwargs
                  }
        publish = farine.amqp.Publisher(self.service, '{}__{}'.format(self.service, self.remote))
        publish.send(message,
                     correlation_id=self.correlation_id,
                     reply_to=self.queue.name,
                     declare=[self.queue],
                    )
        while self.running:
            try:
                self.start(forever=False, timeout=self.timeout)
            except:#pylint:disable=bare-except
                raise exceptions.RPCError(traceback.format_exc())
            # Iterate over the queue
            while not self.results.empty():
                result = self.results.get()
                if result.get('__except__'):
                    raise exceptions.RPCError(result['__except__'])
                elif result.get('__end__'):
                    self.running = False
                elif result.get('body'):
                    yield result['body']

    def __getattr__(self, method, *args, **kwargs):
        self.remote = method
        return self.__wrap_rpc__



#-*- coding:utf-8 -*-
"""
RPC over AMQP implementation: server side.
"""
import types
import traceback
import farine.amqp

class Server(farine.amqp.Consumer):
    """
    RPC Server which is
    an AMQP consumer then producer.
    """
    prefetch_count = 1
    routing_key_format = '{service}__{callback_name}'

    @farine.amqp.publish()
    def main_callback(self, publish, result, message):#pylint:disable=arguments-differ
        message.ack()
        #1. Retrieve the callback result.
        try:
            with self.database():
                result = self.callback(*result['args'], **result['kwargs'])
        except:#pylint:disable=bare-except
            result = {'__except__': traceback.format_exc()}
            publish(result,
                    routing_key=message.properties['reply_to'],
                    correlation_id=message.properties['correlation_id'],
                   )
            return
        #2. Convert the callback result to a generator
        if not isinstance(result, types.GeneratorType):
            result = iter((result,))
        #3. Now iterate over the result until the end.
        # If there is an exception, stop iterating.
        try:
            for res in result:
                publish({'body':res, '__end__': False},
                        routing_key=message.properties['reply_to'],
                        correlation_id=message.properties['correlation_id'],
                       )
            publish({'__end__': True},
                    routing_key=message.properties['reply_to'],
                    correlation_id=message.properties['correlation_id'],
                   )
        except:#pylint:disable=bare-except
            res = {'__except__': traceback.format_exc()}
            publish(res,
                    routing_key=message.properties['reply_to'],
                    correlation_id=message.properties['correlation_id'],
                   )



#-*- coding:utf-8 -*-
"""
The publisher module manage all the producer's messages workflow:
    - Connection to the server
    - Publish the message to an exchange
    - Error handler (reconnection, retry, etc.)
"""
import logging
from kombu import Connection, Exchange
from kombu.pools import producers, connections
import farine.settings

LOGGER = logging.getLogger(__name__)

class Publisher(object):
    """
    Publisher generic class. Also known as Producer.
    """
    auto_delete = False

    def __init__(self, name, routing_key, service=None):
        """
        Messages will be published to `exchange`, using these different settings.
        :param name: The exchange name, required.
        :type name: str
        :param routing_key: The key used to route the message.
        :type routing_key: None, str
        :param service: The service's name. Used to get the configuration
        :type service: None, str
        :rtype: None
        """
        self.settings = getattr(farine.settings, service or name)
        self.routing_key = routing_key
        self.exchange = Exchange(name, type=self.settings['type'],
                                 durable=self.settings['durable'],
                                 auto_delete=self.settings.get('auto_delete', self.auto_delete),
                                 delivery_mode=self.settings['delivery_mode'])
    def get_connection(self):
        """
        Retrieve the connection, lazily.

        :returns: The broker connection.
        :rtype: kombu.connection.Connection
        """
        return Connection(self.settings['amqp_uri'])

    def send(self, message, *args, **kwargs):
        """
        Send the the `message` to the broker.

        :param message: The message to send. Its type depends on the serializer used.
        :type message: object
        :rtype: None
        """
        routing_keys = kwargs.get('routing_key') or self.routing_key
        routing_keys = [routing_keys] if isinstance(routing_keys, basestring) else routing_keys
        correlation_id = kwargs.get('correlation_id', None)
        reply_to = kwargs.get('reply_to', None)
        declare=[self.exchange] + kwargs.get('declare', [])
        conn = self.get_connection()
        with connections[conn].acquire(block=True) as connection:
            self.exchange.maybe_bind(connection)
            #reply_to.maybe_bind(connection)
            #reply_to.declare(True)
            with producers[connection].acquire(block=True) as producer:
                for routing_key in routing_keys:
                    LOGGER.info('Send message %s to exchange %s with routing_key %s reply_to %s correlation_id %s',
                                message, self.exchange.name, routing_key, reply_to, correlation_id)
                    producer.publish(
                        message,
                        exchange=self.exchange,
                        declare=declare,
                        serializer=self.settings['serializer'],
                        routing_key=routing_key,
                        correlation_id=correlation_id,
                        retry=self.settings['retry'],
                        delivery_mode=self.settings['delivery_mode'],
                        reply_to=reply_to,
                        retry_policy=self.settings['retry_policy'])

    def __call__(self, *args, **kwargs):
        return self.send(*args, **kwargs)

    def close(self):
        """
        Release the connection to the broker.

        :rtype: None
        """
        self.get_connection().release()



#-*- coding:utf-8 -*-
import mock
import farine.discovery
import gevent.threadpool
from fixtures import *
from farine.amqp import Consumer

def test_import_services_error():
    with pytest.raises(SystemExit):
        farine.discovery.import_module('nonexistentmodule')

def test_import_services_ok():
    assert farine.discovery.import_module('farine.tests.testfarine') == None

@mock.patch('gevent.threadpool.ThreadPool')
def test_start(channel, consumer_factory):
    consumer = consumer_factory()
    assert farine.discovery.ENTRYPOINTS
    assert farine.discovery.start()
